package lcsb.mapviewer.services.impl;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang3.math.NumberUtils;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.annotation.services.annotators.AnnotatorException;
import lcsb.mapviewer.annotation.services.annotators.HgncAnnotator;
import lcsb.mapviewer.common.IProgressUpdater;
import lcsb.mapviewer.common.Pair;
import lcsb.mapviewer.common.XmlParser;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.graph.DataMining;
import lcsb.mapviewer.model.map.graph.DataMiningSet;
import lcsb.mapviewer.model.map.graph.DataMiningType;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.reaction.ReactionNode;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.persist.dao.map.graph.DataMiningSetDao;
import lcsb.mapviewer.services.interfaces.IDataMiningService;
import lcsb.mapviewer.services.interfaces.ILogService;
import lcsb.mapviewer.services.utils.InvalidDataMiningInputFile;
import lcsb.mapviewer.services.view.DataMiningSetView;

/**
 * Implementation of the service that manages data mining data.
 * 
 * @author Piotr Gawron
 * 
 */
@Transactional(value = "txManager")
public class DataMiningService extends XmlParser implements IDataMiningService {

	/**
	 * Name of the source column (element that misses connection) in the input
	 * file containing information about data mining.
	 */
	private static final String	SOURCE_COLUMN_NAME = "Source";

	/**
	 * Name of the target column (connection that is missing) in the input file.
	 */
	private static final String	TARGET_COLUMN_NAME = "Target";

	/**
	 * Name of the description column in the input file.
	 */
	private static final String	DESC_COLUMN_NAME	 = "Description";

	/**
	 * Name of the column where references are kept in the input file.
	 */
	private static final String	REF_COLUMN_NAME		 = "References";

	/**
	 * Default class logger.
	 */
	private final Logger				logger						 = Logger.getLogger(DataMiningService.class);

	/**
	 * Data access object for {@link DataMiningSet} objects.
	 */
	@Autowired
	private DataMiningSetDao		dataMinigSetDao;

	/**
	 * Class used to work on {@link MiriamType#HGNC} annotations.
	 */
	@Autowired
	private HgncAnnotator				hgncAnnotator;

	/**
	 * Service that allows to log system messages.
	 */
	@Autowired
	private ILogService					logService;

	/**
	 * @return the logService
	 */
	public ILogService getLogService() {
		return logService;
	}

	/**
	 * @param logService
	 *          the logService to set
	 */
	public void setLogService(ILogService logService) {
		this.logService = logService;
	}

	/**
	 * Map that contains cached information if identifiers are hgnc valid names or
	 * not.
	 */
	private Map<String, Boolean> isHgnc = new HashMap<String, Boolean>();

	/**
	 * Returns true if the parameter is valid hgnc identifier.
	 * 
	 * @param string
	 *          name of supposed hgnc name
	 * @return true if paraeter is valid hgnc identifier
	 */
	protected boolean isHgncName(String string) {
		if (string == null) {
			return false;
		}
		Boolean status = isHgnc.get(string);
		if (status == null) {
			try {
				status = hgncAnnotator.isValidHgncMiriam(new MiriamData(MiriamType.HGNC_SYMBOL, string));
			} catch (AnnotatorException e) {
				logger.error(e);
				status = false;
			}
			isHgnc.put(string, status);
		}
		return status;
	}

	@Override
	public Set<DataMining> parseData(String filename, DataMiningType type, Collection<Element> nodes, Set<Reaction> reactions, IProgressUpdater updater)
			throws IOException, InvalidDataMiningInputFile {
		InputStream is = new FileInputStream(filename);
		String name = new File(filename).getName();
		String warningPrefix = "[Data mining, " + name + "] ";

		return parseData(nodes, reactions, is, warningPrefix);
	}

	/**
	 * This method creates set of {@link DataMining} elements from input stream.
	 * 
	 * @param is
	 *          {@link InputStream}
	 * @param nodes
	 *          nodes for which data mining connections shoul be generated
	 * @param reactions
	 *          list of already existing reactions in the model
	 * @param warningPrefix
	 *          prefix used in warnings
	 * @return data mining connections for given data
	 * @throws InvalidDataMiningInputFile
	 *           thrown when the input file is invalid
	 */
	protected Set<DataMining> parseData(Collection<Element> nodes, Set<Reaction> reactions, InputStream is, String warningPrefix)
			throws InvalidDataMiningInputFile {
		Set<DataMining> result = new HashSet<DataMining>();
		String content = null;
		try {
			content = inputStreamToString(is).replaceAll("[\r\n]+", "\n");
		} catch (IOException e) {
			throw new InvalidDataMiningInputFile(warningPrefix + " " + e.getMessage(), e);
		}
		String[] lines = content.split("\n");
		int headerLine = -1;
		for (int i = 0; i < lines.length; i++) {
			if (!lines[i].startsWith("#") && !lines[i].isEmpty()) {
				headerLine = i;
				break;
			}
		}
		if (headerLine < 0) {
			throw new InvalidDataMiningInputFile(warningPrefix + " No header line found");
		}
		String[] columns = lines[headerLine].split("\t");
		Integer sourceColumnId = null;
		Integer targetColumnId = null;
		Integer descColumnId = null;
		Integer refColumnId = null;
		for (int i = 0; i < columns.length; i++) {
			String col = columns[i].trim();
			if (SOURCE_COLUMN_NAME.equalsIgnoreCase(col)) {
				sourceColumnId = i;
			} else if (TARGET_COLUMN_NAME.equalsIgnoreCase(col)) {
				targetColumnId = i;
			} else if (DESC_COLUMN_NAME.equalsIgnoreCase(col)) {
				descColumnId = i;
			} else if (REF_COLUMN_NAME.equalsIgnoreCase(col)) {
				refColumnId = i;
			} else {
				logger.debug(headerLine + ": \"" + lines[headerLine] + "\"");
				throw new InvalidDataMiningInputFile(warningPrefix + " Unknown column name: \"" + col + "\". Only " //
						+ "\"" + SOURCE_COLUMN_NAME + "\", " //
						+ "\"" + TARGET_COLUMN_NAME + "\", " //
						+ "\"" + DESC_COLUMN_NAME + "\", " //
						+ "\"" + REF_COLUMN_NAME + "\" " //
						+ " are available.");
			}
		}
		if (sourceColumnId == null) {
			throw new InvalidDataMiningInputFile(warningPrefix + " Column \"" + SOURCE_COLUMN_NAME + "\" wasn't found in the input file.");
		}

		if (targetColumnId == null) {
			throw new InvalidDataMiningInputFile(warningPrefix + " Column \"" + TARGET_COLUMN_NAME + "\" wasn't found in the input file.");
		}
		Map<MiriamData, List<Element>> map = new HashMap<>();
		for (Element element : nodes) {
			for (MiriamData md : element.getMiriamData()) {
				List<Element> elList = map.get(md);
				if (elList == null) {
					elList = new ArrayList<>();
					map.put(md, elList);
				}
				elList.add(element);
			}
		}
		Set<Pair<MiriamData, MiriamData>> reactionsHash = createReactionHashMap(reactions);

		for (int i = headerLine + 1; i < lines.length; i++) {
			String[] tmp = lines[i].split("\t", -1);
			if (tmp.length != columns.length) {
				String warning = warningPrefix + "Row " + i + " contains invalid number of columns. Skipping... (row content: " + lines[i] + ") ";
				logger.warn(warning);
				continue;
			}
			String id1 = tmp[sourceColumnId];
			String id2 = tmp[targetColumnId];
			MiriamData md1 = parseIdentifier(id1, warningPrefix);
			MiriamData md2 = parseIdentifier(id2, warningPrefix);
			String description = null;
			if (descColumnId != null) {
				description = tmp[descColumnId].trim();
			} else {
				MiriamData hgncSymbol = null;
				try {
					if (md2.getDataType().equals(MiriamType.HGNC_SYMBOL)) {
						hgncSymbol = md2;
					} else if (md2.getDataType().equals(MiriamType.HGNC)) {
						hgncSymbol = hgncAnnotator.hgncIdToHgncName(md2);
					}
				} catch (AnnotatorException e) {
					logger.warn(warningPrefix + " Problem with transfomring identifier to " + MiriamType.HGNC_SYMBOL.getCommonName() + ". Identifier: " + md2, e);
				}
				if (hgncSymbol != null) {
					description = hgncSymbol.getResource();
				}
			}
			List<MiriamData> ref = new ArrayList<MiriamData>();
			if (refColumnId != null) {
				ref.addAll(createReferences(warningPrefix, tmp[refColumnId]));
			}
			if (md1 != null && md2 != null) {
				boolean exists = false;
				if (md1.compareTo(md2) < 0) {
					exists = reactionsHash.contains(new Pair<MiriamData, MiriamData>(md1, md2));
				} else {
					exists = reactionsHash.contains(new Pair<MiriamData, MiriamData>(md2, md1));
				}
				if (!exists) {
					List<Element> listEl = map.get(md1);
					if (listEl != null) {
						for (Element el : listEl) {
							DataMining dm = new DataMining();
							dm.setElement(el);
							dm.setDescription(description);
							dm.addReferences(copyList(ref));
							dm.addSuggestedConnection(new MiriamData(md2));
							dm.setType(DataMiningType.UNKNOWN);
							result.add(dm);
						}
					} else {
						logger.warn(warningPrefix + " Element with identifier \"" + md1.getDataType().getCommonName() + ": " + md1.getResource() + "\" doesn't exist");
					}
				}
			}
		}
		return result;
	}

	/**
	 * Creates list of references from string with comma separated identifiers.
	 * 
	 * @param warningPrefix
	 *          prefix used for warnings
	 * @param refString
	 *          strign with comma separated pubmed identifiers
	 * @return list of references from string with comma separated identifiers
	 */
	public List<MiriamData> createReferences(String warningPrefix, String refString) {
		List<MiriamData> ref = new ArrayList<>();
		for (String string : refString.split(",")) {
			if (!string.isEmpty()) {
				String id = string.replaceAll("urn:miriam:pubmed:", "").trim();
				if (NumberUtils.isDigits(id)) {
					ref.add(new MiriamData(MiriamType.PUBMED, id.trim()));
				} else {
					logger.warn(warningPrefix + "Invalid pubmed identifier: \"" + string);
				}
			}
		}
		return ref;
	}

	/**
	 * Creates list with copies of the {@link MiriamData} objects.
	 * 
	 * @param ref
	 *          list to copy
	 * @return list with copies of the {@link MiriamData} objects
	 */
	private List<MiriamData> copyList(List<MiriamData> ref) {
		List<MiriamData> result = new ArrayList<>();
		for (MiriamData miriamData : ref) {
			result.add(new MiriamData(miriamData));
		}
		return result;
	}

	/**
	 * Parses string identifiers and returns {@link MiriamData} object with
	 * information from the identifier.
	 * 
	 * @param identifier
	 *          identfier to be parsed
	 * @param warningPrefix
	 *          prefix that should be added when reporting warning
	 * @return {@link MiriamData} object obtained from identifier
	 */
	protected MiriamData parseIdentifier(String identifier, String warningPrefix) {
		MiriamData md1 = null;
		try {
			md1 = MiriamType.getMiriamDataFromIdentifier(identifier.trim());
		} catch (InvalidArgumentException e) {
			try {
				md1 = MiriamType.getMiriamByUri(identifier.trim());
			} catch (InvalidArgumentException e1) {
				logger.warn(warningPrefix + "Invalid identifier: " + identifier.trim());
			}
		}
		return md1;
	}

	/**
	 * Creates a set of pairs ( {@link MiriamData}, {@link MiriamData} )
	 * containing information about all connection in the reactions.
	 * 
	 * @param reactions
	 *          input reactions.
	 * @return set of pairs ( {@link MiriamData}, {@link MiriamData} ) containing
	 *         information about all connection in the input reactions
	 */
	protected Set<Pair<MiriamData, MiriamData>> createReactionHashMap(Set<Reaction> reactions) {
		Set<Pair<MiriamData, MiriamData>> reactionsHash = new HashSet<>();
		for (Reaction reaction : reactions) {
			List<ReactionNode> rNodes = reaction.getReactionNodes();
			for (int i = 0; i < rNodes.size(); i++) {
				for (int j = i + 1; j < rNodes.size(); j++) {
					for (MiriamData md1 : rNodes.get(i).getElement().getMiriamData()) {
						for (MiriamData md2 : rNodes.get(j).getElement().getMiriamData()) {
							if (md1.compareTo(md2) < 0) {
								reactionsHash.add(new Pair<MiriamData, MiriamData>(md1, md2));
							} else {
								reactionsHash.add(new Pair<MiriamData, MiriamData>(md2, md1));
							}
						}
					}
				}
			}
		}
		return reactionsHash;
	}

	@Override
	public Set<DataMining> parseData(DataMiningSet dataMiningSet, Collection<Element> nodes, Set<Reaction> reactions, IProgressUpdater secondPartUpdater)
			throws InvalidDataMiningInputFile {
		String name = dataMiningSet.getName();
		String warningPrefix = "[Data mining, " + name + "] ";
		return parseData(nodes, reactions, new ByteArrayInputStream(dataMiningSet.getInputData()), warningPrefix);
	}

	@Override
	public byte[] getInputDataForLayout(DataMiningSetView dms) {
		DataMiningSet dataMinigSet = dataMinigSetDao.getById(dms.getIdObject());
		if (dataMinigSet == null) {
			return null;
		} else {
			return dataMinigSet.getInputData();
		}
	}
}
