package lcsb.mapviewer.services.impl;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;

import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.converter.model.celldesigner.reaction.ReactionLineData;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.compartment.Compartment;
import lcsb.mapviewer.model.map.compartment.PathwayCompartment;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.reaction.Modifier;
import lcsb.mapviewer.model.map.reaction.Product;
import lcsb.mapviewer.model.map.reaction.Reactant;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.reaction.ReactionNode;
import lcsb.mapviewer.model.map.species.AntisenseRna;
import lcsb.mapviewer.model.map.species.Complex;
import lcsb.mapviewer.model.map.species.Degraded;
import lcsb.mapviewer.model.map.species.Drug;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.Gene;
import lcsb.mapviewer.model.map.species.Ion;
import lcsb.mapviewer.model.map.species.Phenotype;
import lcsb.mapviewer.model.map.species.Protein;
import lcsb.mapviewer.model.map.species.Rna;
import lcsb.mapviewer.model.map.species.SimpleMolecule;
import lcsb.mapviewer.model.map.species.Species;
import lcsb.mapviewer.model.map.species.Unknown;
import lcsb.mapviewer.model.map.species.field.AntisenseRnaRegion;
import lcsb.mapviewer.model.map.species.field.ModificationResidue;
import lcsb.mapviewer.model.map.species.field.RnaRegion;
import lcsb.mapviewer.services.interfaces.IExporterService;
import lcsb.mapviewer.services.utils.data.ExportColumn;
import lcsb.mapviewer.services.utils.data.ExportFileType;
import lcsb.mapviewer.services.view.PubmedAnnotatedElementsView;
import lcsb.mapviewer.services.view.PubmedAnnotatedElementsViewFactory;

/**
 * Implementation of {@link IExporterService} interface that allows to export
 * {@link Model}.
 * 
 * @author Piotr Gawron
 * 
 */
public class ExporterService implements IExporterService {

	/**
	 * Default class logger.
	 */
	private static Logger																					 logger											= Logger.getLogger(ExporterService.class);

	/**
	 * List of publications by {@link Model#getId() model id}. This is used for
	 * caching, as obtaining this list on runtime is quite expensive.
	 */
	private static Map<Integer, List<PubmedAnnotatedElementsView>> pubmedPublicationsForModel	= new HashMap<>();

	/**
	 * Factory used to create {@link PubmedAnnotatedElementsView publication
	 * summary} elements.
	 */
	@Autowired
	private PubmedAnnotatedElementsViewFactory										 pubmedAnnotatedElementsViewFactory;

	@Override
	public void exportSpecies(ExporterParameters parameters) throws IOException {
		String content = getExportSpeciesString(parameters);
		writeToFile(content, parameters.getFileName());
	}

	@Override
	public String getExportSpeciesString(ExporterParameters parameters) {
		Set<String> uniqeLines = new HashSet<String>();
		StringBuilder builder = new StringBuilder();
		for (ExportColumn column : parameters.getColumns()) {
			builder.append(column.getTitle() + "\t");
		}
		builder.append("\n");
		for (Model model : parameters.getModels()) {
			for (Element alias : model.getElements()) {
				if (alias instanceof Species) {
					Species speciesAlias = (Species) alias;
					boolean toBeProcessed = false;
					// check type
					for (Class<?> type : parameters.getTypes()) {
						if (type.isAssignableFrom(speciesAlias.getClass())) {
							toBeProcessed = true;
							break;
						}
					}
					if (!toBeProcessed && parameters.getTypes().size() > 0) {
						continue;
					}

					// check where it lays
					toBeProcessed = (parameters.getIncludedAliases().size() == 0);
					for (Compartment ca : parameters.getIncludedAliases()) {
						if (ca.cross(alias)) {
							toBeProcessed = true;
						}
					}
					for (Compartment ca : parameters.getExcludedAliases()) {
						if (ca.cross(alias)) {
							toBeProcessed = false;
						}
					}
					if (toBeProcessed) {
						List<String> elements = getExportStringForOneSpecies(speciesAlias, parameters);
						for (String string : elements) {
							uniqeLines.add(string);
						}
					}
				}
			}
		}
		List<String> sortedElements = new ArrayList<String>();
		sortedElements.addAll(uniqeLines);
		Collections.sort(sortedElements);
		for (String string : sortedElements) {
			builder.append(string);
			builder.append("\n");
		}
		return builder.toString();
	}

	/**
	 * Returns strings with species alias exported using given parameters. Many
	 * lines can be exported because one species alias can belong to many
	 * pathways.
	 * 
	 * @param parameters
	 *          {@link ExporterParameters export parameters}
	 * @param species
	 *          export string are generated for this alias
	 * @return list of strings representing exported data
	 */
	protected List<String> getExportStringForOneSpecies(Species species, ExporterParameters parameters) {
		if (parameters.getFileType() != ExportFileType.TAB_SEPARATED) {
			throw new InvalidArgumentException("Unhandled file type: " + parameters.getFileType());
		}

		List<Compartment> aliases = getComponentList(species);
		List<String> result = new ArrayList<String>();

		Compartment smallestCompartment = getCompartment(species);

		List<ExportColumn> processedColumns = new ArrayList<ExportColumn>();
		for (ExportColumn column : parameters.getColumns()) {
			if (column.getClazz().isAssignableFrom(Species.class)) {
				processedColumns.add(column);
			}
		}

		for (Compartment compartment : aliases) {
			String separator = "";
			StringBuilder line = new StringBuilder();
			for (ExportColumn column : processedColumns) {
				line.append(separator);
				separator = "\t";
				if (column == ExportColumn.COMPARTMENT_NAME) {
					if (smallestCompartment == null) {
						line.append("null");
					} else {
						line.append(smallestCompartment.getName());
					}
				} else if (column == ExportColumn.POSITION_TO_COMPARTMENT) {
					line.append(species.getPositionToCompartment());
				} else if (column == ExportColumn.COMPONENT_NAME) {
					line.append(compartment.getName());
				} else if (column == ExportColumn.NAME) {
					if (species instanceof Complex) {
						if (!parameters.isComplexElementsName()) {
							line.append(species.getName());
						} else {
							line.append(getComplexNameWithElements((Complex) species, parameters));
						}
					} else {
						line.append(species.getName());
					}
				} else if (column == ExportColumn.ID) {
					line.append(getId(species));
				} else if (column == ExportColumn.TYPE) {
					line.append(species.getClass().getSimpleName());
				} else if (column == ExportColumn.SUBMODEL) {
					line.append(species.getModelData().getName());
				} else if (column == ExportColumn.COMPLEX) {
					String complex = null;
					if (species.getComplex() != null) {
						complex = species.getComplex().getName();

						complex += " [id=" + getId(species.getComplex()) + "]";
					}
					line.append(complex);
				} else {
					throw new InvalidArgumentException("Unhandled column type: " + column);
				}
			}
			for (MiriamType mt : parameters.getMiriamTypes()) {
				line.append(separator);
				String types = "";
				for (MiriamData md : species.getMiriamData()) {
					if (md.getDataType().equals(mt)) {
						if (!types.isEmpty()) {
							types += ",";
						}
						types += md.getResource();
					}
				}
				if (types.equals("")) {
					line.append(mt.toString() + ": N/A");
				} else {
					line.append(mt.toString() + ": " + types);
				}
			}
			result.add(line.toString().replaceAll("[\n\r]+", " "));
		}
		return result;
	}

	/**
	 * Returns {@link Compartment} to which element belongs to.
	 * 
	 * @param element
	 *          alias of the element
	 * @return {@link Compartment} to which element belongs to
	 */
	private Compartment getCompartment(Species element) {
		Compartment smallestAlias = null;
		List<Compartment> compartmentAliases = new ArrayList<Compartment>();

		for (Compartment alias : element.getModel().getCompartments()) {
			if (alias.cross(element)) {
				if (!(alias instanceof PathwayCompartment)) {
					compartmentAliases.add(alias);
				}
			}
		}

		if (compartmentAliases.size() > 0) {
			smallestAlias = compartmentAliases.get(0);
		}
		for (Compartment compartmentAlias : compartmentAliases) {
			if (compartmentAlias.getSize() < smallestAlias.getSize()) {
				smallestAlias = compartmentAlias;
			}
		}
		return smallestAlias;
	}

	/**
	 * Returns a list of pathways to which element belongs to.
	 * 
	 * @param element
	 *          alias of the element
	 * @return list of {@link Compartment} representing pathways
	 */
	private List<Compartment> getComponentList(Species element) {
		Model model = element.getModel();

		List<Compartment> aliases = new ArrayList<Compartment>();
		for (Compartment alias : model.getCompartments()) {
			if (alias instanceof PathwayCompartment) {
				if (alias.cross(element)) {
					aliases.add(alias);
				}
			}
		}

		// in case the element is outside any component then we have to crate
		// artifical null component
		if (aliases.size() == 0) {
			PathwayCompartment nullAlias = new PathwayCompartment("default");
			nullAlias.setName("null");
			aliases.add(nullAlias);
		}

		return aliases;
	}

	/**
	 * Returns name of a complex with the names of all species included in this
	 * complex.
	 * 
	 * @param complex
	 *          complex for which name is computed
	 * @param parameters
	 *          parameter limits the species that will be included in generated
	 *          name
	 * @return name of a complex with the names of all species included in this
	 *         complex
	 */
	private String getComplexNameWithElements(Complex complex, ExporterParameters parameters) {
		StringBuilder result = new StringBuilder();
		String separator = "";
		for (Element species2 : complex.getAllSimpleChildren()) {
			boolean add = false;
			if (parameters.getTypes().size() == 0) {
				add = true;
			}
			for (Class<?> class1 : parameters.getTypes()) {
				if (class1.isAssignableFrom(species2.getClass())) {
					add = true;
				}
			}
			if (add) {
				result.append(separator);
				result.append(species2.getName());
				separator = ":";
			}

		}
		if (result.length() == 0) {
			return complex.getName();
		} else {
			return result.toString();
		}
	}

	/**
	 * Writes content into a file.
	 * 
	 * @param content
	 *          content that should be written into the file.
	 * @param fileName
	 *          file name
	 * @throws IOException
	 *           thrown when there is a problem with accessing file
	 */
	protected void writeToFile(String content, String fileName) throws IOException {
		logger.debug("Save file: " + fileName);
		File file = new File(fileName);

		// if file doesnt exists, then create it
		if (!file.exists()) {
			// first create directory structure
			file.getParentFile().mkdirs();
			file.createNewFile();
		}

		FileWriter fw = new FileWriter(file.getAbsoluteFile());
		BufferedWriter bw = new BufferedWriter(fw);
		bw.write(content);
		bw.close();
	}

	@Override
	public String getExportInteractionString(ExporterParameters parameters) {
		Set<String> uniqueLines = new HashSet<String>();
		if (parameters.getFileType() != ExportFileType.SIF && parameters.getFileType() != ExportFileType.TAB_SEPARATED) {
			throw new InvalidArgumentException("Unsupported file format: " + parameters.getFileType());
		}

		for (Model model : parameters.getModels()) {
			for (Reaction reaction : model.getReactions()) {
				int counter = 0;
				for (ReactionNode node : reaction.getReactionNodes()) {
					Element element = node.getElement();
					for (Class<?> type : parameters.getTypes()) {
						if (type.isAssignableFrom(element.getClass())) {
							counter++;
							break;
						}
					}
				}
				if (counter > 1) {
					uniqueLines.addAll(getExportSingleInteractionString(reaction, parameters));
				}
			}
		}
		if (parameters.isMoleculeEdges()) {
			for (Model model : parameters.getModels()) {
				for (Element element : model.getElements()) {
					if (element instanceof Species) {
						if (element instanceof SimpleMolecule) {
							boolean ok = false;
							for (Compartment alias : parameters.getIncludedAliases()) {
								if (alias.cross(element)) {
									ok = true;
								}
							}
							for (Compartment alias : parameters.getExcludedAliases()) {
								if (alias.cross(element)) {
									ok = false;
								}
							}
							if (ok) {
								String string = getExportSingleInteractionStringFromSpecies((SimpleMolecule) element, parameters);
								if (parameters.getFileType() == ExportFileType.SIF) {
									if (!uniqueLines.contains(string.replace("molecule_type", "reaction_type"))) {
										uniqueLines.add(string);
									}
								} else {
									uniqueLines.add(string);
								}
							}
						}
					}
				}
			}
		}
		List<String> sortedLines = new ArrayList<String>();
		sortedLines.addAll(uniqueLines);
		Collections.sort(sortedLines);
		StringBuilder result = new StringBuilder();
		for (String string : sortedLines) {
			if (!string.equals("")) {
				result.append(string);
				result.append("\n");
			}
		}

		return result.toString();
	}

	/**
	 * Creates interaction from molecule (molecule is used as extension of
	 * reaction between different types of elements).
	 * 
	 * @param species
	 *          molecule which is treaded as a reaction element
	 * @param parameters
	 *          {@link ExporterParameters export parameters}
	 * @return string representing exported reaction
	 */
	protected String getExportSingleInteractionStringFromSpecies(SimpleMolecule species, ExporterParameters parameters) {
		if (parameters.getFileType() != ExportFileType.SIF) {
			throw new InvalidArgumentException("Unsupported file format: " + parameters.getFileType());
		}
		StringBuilder result = new StringBuilder();
		Set<String> elements = new HashSet<String>();
		for (Model model : parameters.getModels()) {
			for (Reaction reaction : model.getReactions()) {
				boolean goodReaction = false;
				for (ReactionNode node : reaction.getReactionNodes()) {
					if (node.getElement() instanceof SimpleMolecule) {
						Element species2 = node.getElement();
						if (species.getName().equalsIgnoreCase(species2.getName())) {
							goodReaction = true;
						}
					}
				}
				if (!goodReaction) {
					continue;
				}
				for (ReactionNode node : reaction.getReactionNodes()) {
					boolean ok = false;
					if (parameters.getIncludedAliases().size() == 0) {
						ok = true;
					}
					for (Compartment alias2 : parameters.getIncludedAliases()) {
						if (alias2.cross(node.getElement())) {
							ok = true;
							break;
						}
					}
					for (Compartment alias2 : parameters.getExcludedAliases()) {
						if (alias2.cross(node.getElement())) {
							ok = false;
							break;
						}
					}
					if (ok) {
						Element element = node.getElement();
						for (Class<?> type : parameters.getTypes()) {
							if (type.isAssignableFrom(element.getClass())) {
								if (element instanceof Complex) {
									if (!parameters.isComplexElementsName()) {
										elements.add(((Species) element).getName());
									} else {
										String name = getComplexNameWithElements((Complex) element, parameters);
										elements.add(name);
									}
								} else {
									elements.add(((Species) element).getName());
								}
								break;
							}
						}
					}
				}
			}
		}
		if (elements.size() <= 1) {
			return "";
		}
		List<String> list = new ArrayList<String>();
		list.addAll(elements);
		result.append(list.get(0));
		result.append("\tmolecule_type");
		for (int i = 1; i < list.size(); i++) {
			result.append("\t");
			result.append(list.get(i));
		}
		return result.toString();

	}

	/**
	 * Returns strings with reactions exported using given parameters. Many lines
	 * can be exported because one species alias can belong to many pathways.
	 * 
	 * @param parameters
	 *          {@link ExporterParameters export parameters}
	 * @param reaction
	 *          export string are generated for this reaction
	 * @return list of strings representing exported data
	 */
	protected List<String> getExportSingleInteractionString(Reaction reaction, ExporterParameters parameters) {
		List<String> result = new ArrayList<String>();
		if (parameters.getFileType() == ExportFileType.SIF) {
			StringBuilder builder = new StringBuilder();
			Set<Element> elements = new HashSet<>();
			for (ReactionNode node : reaction.getReactionNodes()) {
				boolean ok = false;
				if (parameters.getIncludedAliases().size() == 0) {
					ok = true;
				}
				for (Compartment alias2 : parameters.getIncludedAliases()) {
					if (alias2.cross(node.getElement())) {
						ok = true;
						break;
					}
				}
				for (Compartment alias2 : parameters.getExcludedAliases()) {
					if (alias2.cross(node.getElement())) {
						ok = false;
						break;
					}
				}
				if (ok) {
					Element element = node.getElement();
					if (parameters.getTypes().size() == 0) {
						elements.add(element);
					} else {
						for (Class<?> type : parameters.getTypes()) {
							if (type.isAssignableFrom(element.getClass())) {
								elements.add(element);
								break;
							}
						}
					}
				}
			}
			Set<String> elementsNames = new HashSet<>();
			for (Element element : elements) {
				if (element instanceof Complex) {
					if (!parameters.isComplexElementsName()) {
						elementsNames.add(((Species) element).getName());
					} else {
						String name = getComplexNameWithElements((Complex) element, parameters);
						elementsNames.add(name);
					}

				} else {
					elementsNames.add(((Species) element).getName());
				}
			}

			if (elementsNames.size() <= 1) {
				return result;
			}
			List<String> list = new ArrayList<String>();
			list.addAll(elementsNames);
			builder.append(list.get(0));
			builder.append("\treaction_type");
			for (int i = 1; i < list.size(); i++) {
				builder.append("\t");
				builder.append(list.get(i));
			}
			result.add(builder.toString());
		} else if (parameters.getFileType() == ExportFileType.TAB_SEPARATED) {
			if (parameters.getColumns() == null || parameters.getColumns().size() == 0) {
				result.add(getOldFormatString(reaction, parameters));
			} else {
				result.addAll(getNewFormatString(reaction, parameters));
			}
		} else {
			throw new InvalidArgumentException("Unsupported file format: " + parameters.getFileType());
		}
		return result;
	}

	/**
	 * Creates tab separated string for reaction in old format...
	 * 
	 * @param reaction
	 *          reaction to transform
	 * @param parameters
	 *          {@link ExporterParameters export parameters}
	 * @return export string representation of the reaction
	 */
	private String getOldFormatString(Reaction reaction, ExporterParameters parameters) {
		StringBuilder result = new StringBuilder();
		result.append(reaction.getIdReaction());
		result.append("\t");
		result.append(ReactionLineData.getByReactionType(reaction.getClass()));
		result.append("\t");
		result.append("reversible=" + reaction.isReversible());
		result.append("\t");
		result.append("kineticLaw=" + reaction.isKineticLaw());
		result.append("\t");
		if (reaction.getMiriamData().size() > 0) {
			result.append("REACTION_MIRIAM_DATA");
			result.append("\t");
			for (MiriamData md : reaction.getMiriamData()) {
				result.append("data_type=" + md.getDataType().getUris().get(0));
				result.append("\t");
				result.append("relation_type=" + md.getRelationType().getStringRepresentation());
				result.append("\t");
				result.append("resource=" + md.getResource());
				result.append("\t");
			}
		}
		for (Reactant reactant : reaction.getReactants()) {
			result.append("REACTANT");
			result.append("\t");
			result.append(getExportSingleSpecies((Species) reactant.getElement(), parameters));
			result.append("\t");
		}
		for (Product product : reaction.getProducts()) {
			result.append("PRODUCT");
			result.append("\t");
			result.append(getExportSingleSpecies((Species) product.getElement(), parameters));
			result.append("\t");
		}
		for (Modifier modifier : reaction.getModifiers()) {
			result.append("MODIFIER");
			result.append("\t");
			result.append("type=" + modifier.getClass().getSimpleName());
			result.append("\t");
			result.append(getExportSingleSpecies((Species) modifier.getElement(), parameters));
			result.append("\t");
		}
		return result.toString();
	}

	/**
	 * Creates tab separated string for reaction in new format...
	 * 
	 * @param reaction
	 *          reaction to transform
	 * @param parameters
	 *          {@link ExporterParameters export parameters}
	 * @return export string representation of the reaction
	 */
	private List<String> getNewFormatString(Reaction reaction, ExporterParameters parameters) {
		List<String> result = new ArrayList<String>();
		StringBuilder builder = new StringBuilder();
		boolean splitReaction = false;
		for (ExportColumn column : parameters.getColumns()) {
			if (column.getClazz().isAssignableFrom(Reaction.class)) {
				if (column.equals(ExportColumn.REACTION_ID)) {
					builder.append(reaction.getIdReaction());
				} else if (column.equals(ExportColumn.REACTION_REVERSIBLE)) {
					builder.append(reaction.isReversible());
				} else if (column.equals(ExportColumn.REACTION_TYPE)) {
					builder.append(ReactionLineData.getByReactionType(reaction.getClass()));
				} else if (column.equals(ExportColumn.REACTION_TWO_ELEMENT)) {
					splitReaction = true;
					continue;
				} else {
					throw new InvalidArgumentException("Unknown column type: " + column);
				}
			}
			builder.append("\t");
		}
		for (MiriamType mt : parameters.getMiriamTypes()) {
			String types = "";
			for (MiriamData md : reaction.getMiriamData()) {
				if (md.getDataType().equals(mt)) {
					if (!types.isEmpty()) {
						types += ",";
					}
					types += md.getResource();
				}
			}
			if (types.equals("")) {
				builder.append(mt.toString() + ": N/A");
			} else {
				builder.append(mt.toString() + ": " + types);
			}
			builder.append("\t");
		}

		int counter = 0;

		if (splitReaction) {
			String base = builder.toString();
			Set<Element> inputs = new HashSet<Element>();
			Set<Element> outputs = new HashSet<Element>();
			for (Reactant reactant : reaction.getReactants()) {
				if (!isAssignable(parameters, reactant.getElement())) {
					continue;
				}
				if (reactant.getElement() instanceof Complex) {
					inputs.addAll(((Complex) (reactant.getElement())).getAllChildren());
				} else {
					inputs.add(reactant.getElement());
				}
			}
			for (Product product : reaction.getProducts()) {
				if (!isAssignable(parameters, product.getElement())) {
					continue;
				}
				if (product.getElement() instanceof Complex) {
					outputs.addAll(((Complex) (product.getElement())).getAllChildren());
				} else {
					outputs.add(product.getElement());
				}
			}
			for (Modifier modifier : reaction.getModifiers()) {
				if (!isAssignable(parameters, modifier.getElement())) {
					continue;
				}
				if (modifier.getElement() instanceof Complex) {
					inputs.addAll(((Complex) (modifier.getElement())).getAllChildren());
				} else {
					inputs.add(modifier.getElement());
				}
			}
			for (Element input : inputs) {
				if (!isAssignable(parameters, input)) {
					continue;
				}
				for (Element output : outputs) {
					if (!isAssignable(parameters, output)) {
						continue;
					}
					String line = base + "\t" + getId(input) + "\t" + getId(output);
					result.add(line);
				}
			}
		} else {

			for (Reactant reactant : reaction.getReactants()) {
				if (!isAssignable(parameters, reactant.getElement())) {
					continue;
				}
				counter++;

				builder.append("REACTANT");
				builder.append("\t");
				builder.append(getId(reactant.getElement()));
				builder.append("\t");
			}
			for (Product product : reaction.getProducts()) {
				if (!isAssignable(parameters, product.getElement())) {
					continue;
				}
				counter++;

				builder.append("PRODUCT");
				builder.append("\t");
				builder.append(getId(product.getElement()));
				builder.append("\t");
			}
			for (Modifier modifier : reaction.getModifiers()) {
				if (!isAssignable(parameters, modifier.getElement())) {
					continue;
				}
				counter++;

				builder.append("MODIFIER");
				builder.append("\t");
				builder.append(getId(modifier.getElement()));
				builder.append("\t");
			}
			if (counter > 1) {
				result.add(builder.toString());
			}
		}
		return result;
	}

	/**
	 * Checks if element is acceptable by {@link ExporterParameters export
	 * parameters}.
	 * 
	 * @param parameters
	 *          {@link ExporterParameters export parameters}
	 * @param element
	 *          element to check
	 * @return true if element can exist in export string, <code>false</code>
	 *         otherwise
	 */
	private boolean isAssignable(ExporterParameters parameters, Element element) {
		boolean ok = false;
		for (Class<?> clazz : parameters.getTypes()) {
			if (clazz.isAssignableFrom(element.getClass())) {
				ok = true;
			}
		}
		return ok;
	}

	/**
	 * Creates tab separated string for species for full mode.
	 * 
	 * @param alias
	 *          alias of the element to transform
	 * @param parameters
	 *          {@link ExporterParameters export parameters}
	 * @return export string representation of the element
	 */
	private String getExportSingleSpecies(Species alias, ExporterParameters parameters) {
		if (parameters.getFileType() != ExportFileType.TAB_SEPARATED) {
			throw new InvalidArgumentException("Unsupported file format: " + parameters.getFileType());
		}
		StringBuilder sb = new StringBuilder();
		sb.append(alias.getClass().getSimpleName() + "\t");
		sb.append("id=" + getId(alias) + "\t");
		sb.append("name=" + alias.getName() + "\t");
		sb.append("fullName=" + alias.getFullName() + "\t");
		sb.append("charge=" + alias.getCharge() + "\t");
		sb.append("initialAmount=" + alias.getInitialAmount() + "\t");
		sb.append("initialConcentration=" + alias.getInitialConcentration() + "\t");
		String compartment = null;
		if (alias.getCompartment() != null) {
			compartment = alias.getCompartment().getName();
		}
		sb.append("compartment=" + compartment);
		sb.append("\t");
		String complexId = null;
		if (alias.getComplex() != null) {
			complexId = getId(alias.getComplex());
		}
		sb.append("complex=" + complexId);
		sb.append("\t");
		sb.append("hasOnlySubstanceUnits=" + alias.hasOnlySubstanceUnits());
		sb.append("\t");
		sb.append("positionToCompartment=" + alias.getPositionToCompartment());
		sb.append("\t");

		if (alias instanceof Protein) {
			Protein protein = (Protein) alias;
			sb.append("dimer=" + protein.getHomodimer());
			sb.append("\t");
			sb.append("structuralState=" + protein.getStructuralState());
			sb.append("\t");
			sb.append("hypothetical=" + protein.isHypothetical());
			sb.append("\t");
			for (ModificationResidue mr : protein.getModificationResidues()) {
				sb.append("MODIFICATION RESIDUE");
				sb.append("\t");
				sb.append("angle=" + mr.getAngle());
				sb.append("\t");
				sb.append("name=" + mr.getName());
				if (mr.getState() != null) {
					sb.append("\t");
					sb.append("shortName=" + mr.getState().getAbbreviation());
				}
				sb.append("\t");
				sb.append("side=" + mr.getSide());
				sb.append("\t");
				sb.append("size=" + mr.getSize());
				sb.append("\t");
				if (mr.getState() != null) {
					sb.append("state=" + mr.getState().getFullName() + "\t");
				}
			}
		} else if (alias instanceof Gene) {
			Gene gene = (Gene) alias;
			for (ModificationResidue mr : gene.getModificationResidues()) {
				sb.append("MODIFICATION RESIDUE");
				sb.append("\t");
				sb.append("angle=" + mr.getAngle());
				sb.append("\t");
				sb.append("name=" + mr.getName());
				sb.append("\t");
				if (mr.getState() != null) {
					sb.append("shortName=" + mr.getState().getAbbreviation() + "\t");
				}
				sb.append("side=" + mr.getSide());
				sb.append("\t");
				sb.append("size=" + mr.getSize());
				sb.append("\t");
				if (mr.getState() != null) {
					sb.append("state=" + mr.getState().getFullName() + "\t");
				}
			}
		} else if (alias instanceof Complex) {
			Complex complex = (Complex) alias;
			sb.append("dimer=" + complex.getHomodimer());
			sb.append("\t");
			sb.append("hypothetical=" + complex.isHypothetical());
			sb.append("\t");
		} else if (alias instanceof SimpleMolecule) {
			SimpleMolecule molecule = (SimpleMolecule) alias;
			sb.append("dimer=" + molecule.getHomodimer());
			sb.append("\t");
		} else if (alias instanceof Ion) {
			sb.append("");
		} else if (alias instanceof Phenotype) {
			sb.append("");
		} else if (alias instanceof Drug) {
			sb.append("");

		} else if (alias instanceof Rna) {
			Rna rna = (Rna) alias;
			for (RnaRegion region : rna.getRegions()) {
				sb.append("RNA_REGION");
				sb.append("\t");
				sb.append("type=" + region.getType());
				sb.append("\t");
				if (region.getState() != null) {
					sb.append("state=" + region.getState().getFullName() + "\t");
				}
				sb.append("\t");
				sb.append("size=" + region.getSize());
				sb.append("\t");
				sb.append("pos=" + region.getPos());
				sb.append("\t");
			}
		} else if (alias instanceof AntisenseRna) {
			AntisenseRna antisenseRna = (AntisenseRna) alias;
			for (AntisenseRnaRegion region : antisenseRna.getRegions()) {
				sb.append("RNA_REGION" + "\t");
				if (region.getType() != null) {
					sb.append("type=" + region.getType().getName() + "\t");
				}
				sb.append("name=" + region.getName() + "\t");
				sb.append("size=" + region.getSize() + "\t");
				sb.append("pos=" + region.getPos() + "\t");
			}
		} else if (alias instanceof Unknown) {
			sb.append("");
		} else if (alias instanceof Degraded) {
			sb.append("");
		} else {
			throw new InvalidArgumentException("Invalid class type: " + alias.getClass());
		}
		if (alias.getMiriamData().size() > 0) {
			sb.append("MIRIAM_DATA");
			sb.append("\t");
			for (MiriamData md : alias.getMiriamData()) {
				sb.append("data_type=" + md.getDataType().getUris().get(0));
				sb.append("\t");
				sb.append("relation_type=" + md.getRelationType().getStringRepresentation());
				sb.append("\t");
				sb.append("resource=" + md.getResource());
				sb.append("\t");
			}
		}
		return sb.toString();
	}

	@Override
	public void exportReactions(ExporterParameters parameters) throws IOException {
		String content = getExportInteractionString(parameters);
		writeToFile(content, parameters.getFileName());
	}

	@Override
	public void exportCompartments(ExporterParameters parameters) throws IOException {
		String content = getExportCompartmentsString(parameters);
		writeToFile(content, parameters.getFileName());
	}

	@Override
	public String getExportCompartmentsString(ExporterParameters parameters) {
		Set<String> uniqeLines = new HashSet<>();
		StringBuilder builder = new StringBuilder();
		for (Model model : parameters.getModels()) {
			for (Compartment alias : model.getCompartments()) {

				// check where it lays
				boolean toBeProcessed = (parameters.getIncludedAliases().size() == 0);
				for (Compartment ca : parameters.getIncludedAliases()) {
					if (ca.cross(alias)) {
						toBeProcessed = true;
					}
				}
				for (Compartment ca : parameters.getExcludedAliases()) {
					if (ca.cross(alias)) {
						toBeProcessed = false;
					}
				}
				if (toBeProcessed) {
					List<String> elements = getExportStringForOneCompartment(alias, parameters);
					for (String string : elements) {
						uniqeLines.add(string);
					}
				}
			}
		}
		List<String> sortedElements = new ArrayList<>();
		sortedElements.addAll(uniqeLines);
		Collections.sort(sortedElements);
		for (String string : sortedElements) {
			builder.append(string);
			builder.append("\n");
		}
		return builder.toString();
	}

	/**
	 * Exports compartments into set of text lines defining a compartment.
	 * 
	 * @param compartment
	 *          alias of the compartment to export
	 * @param parameters
	 *          parameters used in export
	 * @return string srepresenting compartment in export format
	 */
	private List<String> getExportStringForOneCompartment(Compartment compartment, ExporterParameters parameters) {
		List<String> result = new ArrayList<String>();
		if (parameters.getFileType() != ExportFileType.TAB_SEPARATED) {
			throw new InvalidArgumentException("Unhandled file type: " + parameters.getFileType());
		}
		StringBuilder sb = new StringBuilder();
		sb.append(getId(compartment) + "\t");
		sb.append("name=" + compartment.getName() + "\t");
		sb.append("fullname=" + compartment.getFullName() + "\t");
		if (compartment.getMiriamData().size() > 0) {
			sb.append("MIRIAM_DATA");
			sb.append("\t");
			for (MiriamData md : compartment.getMiriamData()) {
				sb.append("data_type=" + md.getDataType().getUris().get(0));
				sb.append("\t");
				sb.append("relation_type=" + md.getRelationType().getStringRepresentation());
				sb.append("\t");
				sb.append("resource=" + md.getResource());
				sb.append("\t");
			}
		}
		result.add(sb.toString());
		return result;
	}

	/**
	 * Returns identifier that should be used for a given alias.
	 * 
	 * @param alias
	 *          alis for which identifier will be returned
	 * @return identifier that should be used for a given alias
	 */
	private String getId(Element alias) {
		return alias.getElementId();
	}

	@Override
	public List<PubmedAnnotatedElementsView> getPublicationModelSummary(Model model) {
		List<PubmedAnnotatedElementsView> result = pubmedPublicationsForModel.get(model.getId());
		if (result != null) {
			return result;
		} else {
			result = new ArrayList<>();
			pubmedPublicationsForModel.put(model.getId(), result);
		}
		List<Model> models = new ArrayList<>();
		models.add(model);
		models.addAll(model.getSubmodels());
		Set<MiriamData> miriamDatas = new HashSet<>();
		for (Model m : models) {
			for (Element element : m.getElements()) {
				for (MiriamData md : element.getMiriamData()) {
					if (MiriamType.PUBMED.equals(md.getDataType())) {
						miriamDatas.add(md);
					}
				}
			}
			for (Reaction reaction : m.getReactions()) {
				for (MiriamData md : reaction.getMiriamData()) {
					if (MiriamType.PUBMED.equals(md.getDataType())) {
						miriamDatas.add(md);
					}
				}
			}
		}
		for (MiriamData miriamData : miriamDatas) {
			result.add(pubmedAnnotatedElementsViewFactory.create(miriamData, model));
		}
		return result;
	}

}
