package lcsb.mapviewer.services.interfaces;

import java.util.List;

import org.primefaces.model.TreeNode;

import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.model.user.UserAnnotationSchema;
import lcsb.mapviewer.model.user.UserAnnotatorsParam;
import lcsb.mapviewer.services.SecurityException;
import lcsb.mapviewer.services.UserAccessException;
import lcsb.mapviewer.services.utils.CreateProjectParams;
import lcsb.mapviewer.services.view.AuthenticationToken;
import lcsb.mapviewer.services.view.ProjectView;

/**
 * Service that manages projects.
 * 
 * @author Piotr Gawron
 * 
 */
public interface IProjectService {

	/**
	 * Returns a project with a give {@link Project#projectId project identifier}.
	 * 
	 * @param projectId
	 *          {@link Project#projectId project identifier}
	 * @return project with an identifier given as parameter. Null if such project
	 *         doesn't exist.
	 */
	Project getProjectByProjectId(String projectId, AuthenticationToken token) throws UserAccessException;

	/**
	 * Checks if project with a given {@link Project#projectId identifier} exists.
	 * 
	 * @param projectId
	 *          {@link Project#projectId project identifier}
	 * @return <code>true</code> if the project with the given name exists,
	 *         <code>false</code> otherwise
	 */
	boolean projectExists(String projectId);

	/**
	 * Returns list of all projects.
	 * 
	 * @return list of all projects.
	 */
	List<Project> getAllProjects(AuthenticationToken token);

	/**
	 * Returns a list of Project views for all projects.
	 * 
	 * @return list of Project views for all projects
	 */
	List<ProjectView> getAllProjectViews(AuthenticationToken token);

	/**
	 * Updates project.
	 * 
	 * @param projectView
	 *          project view with the data to update
	 */
	void updateProject(ProjectView projectView);

	/**
	 * Removes project from the system.
	 * 
	 * @param homeDir
	 *          directory where the system is deployed
	 * 
	 * @param projectView
	 *          information about project to remove
	 * @param async
	 *          should the operation be done asynchronously
	 * @throws UserAccessException 
	 */
	void removeProject(ProjectView projectView, String homeDir, boolean async, AuthenticationToken token) throws UserAccessException;

	/**
	 * Removes project from the system.
	 * 
	 * @param homeDir
	 *          directory where the system is deployed
	 * @param project
	 *          object to remove
	 * @param async
	 *          should the operation be done asynchronously
	 * @throws UserAccessException 
	 */
	void removeProject(Project project, String homeDir, boolean async, AuthenticationToken token) throws UserAccessException;

	/**
	 * Gets view of the project by the {@link Project#projectId project
	 * identifier}.
	 * 
	 * @param projectId
	 *          {@link Project#projectId project identifier}
	 * @return view of the project
	 * @throws UserAccessException
	 */
	ProjectView getProjectViewByProjectId(String projectId, AuthenticationToken token) throws UserAccessException;

	/**
	 * Adds project to the system.
	 * 
	 * @param project
	 *          object to add
	 */
	void addProject(Project project);

	/**
	 * Return the view of the project identified by database identifier.
	 * 
	 * @param id
	 *          database identifier of the project
	 * @return project view
	 * @throws UserAccessException
	 */
	ProjectView getProjectViewById(Integer id, AuthenticationToken token) throws UserAccessException;

	/**
	 * Creates project using give parameters. See {@link CreateProjectParams}.
	 * 
	 * @param params
	 *          information about project to create
	 * @throws SecurityException 
	 */
	void createProject(CreateProjectParams params) throws SecurityException;

	/**
	 * Creates empty {@link ProjectView}.
	 * 
	 * @return empty {@link ProjectView}
	 */
	ProjectView createEmpty();

	/**
	 * Creates {@link TreeNode} that contains information about all classes that
	 * can be annotated and associate set of
	 * {@link lcsb.mapviewer.annotation.services.annotators.ElementAnnotator
	 * ElementAnnotator} valid for these classes.
	 * 
	 * @param user
	 *          for which information should be retrieved
	 * @return {@link TreeNode} that contains information about all classes that
	 *         can be annotated
	 */
	TreeNode createClassAnnotatorTree(User user);

	/**
	 * Update default set of annotators, lists of required/valid annotations for a
	 * given user and default params used when creating project.
	 * 
	 * @param user
	 *          user for which we updte information
	 * @param sbgnFormat
	 *          new
	 *          {@link lcsb.mapviewer.model.user.UserAnnotationSchema#sbgnFormat}
	 *          value
	 * @param networkLayoutAsDefault
	 *          new
	 *          {@link lcsb.mapviewer.model.user.UserAnnotationSchema#networkLayoutAsDefault}
	 *          value
	 * @param annotatorsTree
	 *          {@link TreeNode} that contains information about all classes that
	 *          can be annotated
	 */
	void updateClassAnnotatorTreeForUser(User user, TreeNode annotatorsTree, boolean sbgnFormat, boolean networkLayoutAsDefault);

	/**
	 * Method that updates information about raw {@link Project} object.
	 * 
	 * @param project
	 *          project to update
	 */
	void updateProject(Project project);

	/**
	 * Returns content of the file that was used to create project given in the
	 * parameter.
	 * 
	 * @param projectView
	 *          view of the {@link Project} for which we are looking for an
	 *          original source file
	 * @return content of the file that was used to create project given in the
	 *         parameter
	 */
	byte[] getInputDataForProject(ProjectView projectView);

	UserAnnotationSchema prepareUserAnnotationSchema(User user);
	
	/**
	 * Retrieves list of annotators {@link UserAnnotatorsParam parameters}
	 * @return
	 * 	list of annotators parameters
	 */
	List<UserAnnotatorsParam> getAnnotatorsParams(User user);

}
