package lcsb.mapviewer.services.search.data;

import java.io.Serializable;

import org.apache.log4j.Logger;

import lcsb.mapviewer.model.map.species.Element;

/**
 * Light class identifing element to visualize on map in a browser.
 * 
 * @author Piotr Gawron
 *
 */
public class ElementIdentifier implements Serializable {
	/**
	 * Default class logger.
	 */
	private static Logger logger = Logger.getLogger(ElementIdentifier.class);

	/**
	 * Enum identifing type of {@link ElementIdentifier}.
	 * 
	 * @author Piotr Gawron
	 *
	 */
	public enum ElementIdentifierType {

		/**
		 * Type representing alias object on the map.
		 */
		ALIAS("ALIAS"),

		/**
		 * Type representing any point on the map.
		 */
		POINT("POINT"),

		/**
		 * Type representing reaction object on the map.
		 */
		REACTION("REACTION");

		/**
		 * String used in Javascript code to identify type.
		 */
		private String jsName;

		/**
		 * Constructor intializing enum type.
		 * 
		 * @param jsName
		 *          {@link #jsName}
		 */
		ElementIdentifierType(String jsName) {
			this.jsName = jsName;
		}

		/**
		 * @return the jsName
		 * @see #jsName
		 */
		public String getJsName() {
			return jsName;
		}

		/**
		 * Returns {@link ElementIdentifierType} by the javascript name.
		 * 
		 * @param jsName
		 *          javascript name of the type
		 * @return {@link ElementIdentifierType} by the javascript name
		 */
		public static ElementIdentifierType getTypeByJsName(String jsName) {
			for (ElementIdentifierType value : values()) {
				if (value.getJsName().equalsIgnoreCase(jsName)) {
					return value;
				}
			}
			logger.warn("Cannot find ElementIdentifierType for jsName: " + jsName);
			return null;
		}

	}

	/**
	 * 
	 */
	private static final long	serialVersionUID = 1L;

	/**
	 * Identifier of the element.
	 */
	private String						objectId;

	/**
	 * Model on which element is placed.
	 */
	private Integer						modelId;

	/**
	 * Type of the element.
	 */
	private String						type;

	/**
	 * Icon associated with the element.
	 */
	private String						icon;

	/**
	 * Default constructor.
	 * 
	 * @param objectId
	 *          {@link #objectId}
	 * @param modelId
	 *          {@link #modelId}
	 * @param type
	 *          {@link #type}
	 * @param icon
	 *          {@link #icon}
	 */
	public ElementIdentifier(Integer objectId, Integer modelId, ElementIdentifierType type, String icon) {
		this(objectId + "", modelId, type, icon);
	}

	/**
	 * Default constructor.
	 * 
	 * @param objectId
	 *          {@link #objectId}
	 * @param modelId
	 *          {@link #modelId}
	 * @param type
	 *          {@link #type}
	 * @param icon
	 *          {@link #icon}
	 */
	public ElementIdentifier(String objectId, Integer modelId, ElementIdentifierType type, String icon) {
		this.objectId = objectId;
		this.modelId = modelId;
		this.type = type.getJsName();
		this.icon = icon;
	}

	/**
	 * Constructor creating object from {@link Element}.
	 * 
	 * @param speciesAlias
	 *          {@link Element} from which {@link ElementIdentifier} is created
	 * @param icon
	 *          {@link #icon}
	 */
	public ElementIdentifier(Element speciesAlias, String icon) {
		this(speciesAlias.getId(), speciesAlias.getModel().getId(), ElementIdentifierType.ALIAS, icon);
	}

	/**
	 * Default constructor.
	 * 
	 * @param objectId
	 *          {@link #objectId}
	 * @param modelId
	 *          {@link #modelId}
	 * @param type
	 *          {@link #type}
	 * @param icon
	 *          {@link #icon}
	 */
	public ElementIdentifier(String objectId, String modelId, ElementIdentifierType type, String icon) {
		this(objectId + "", Integer.valueOf(modelId), type, icon);
	}

	/**
	 * @return the objectId
	 * @see #objectId
	 */
	public String getObjectId() {
		return objectId;
	}

	/**
	 * @param objectId
	 *          the objectId to set
	 * @see #objectId
	 */
	public void setObjectId(String objectId) {
		this.objectId = objectId;
	}

	/**
	 * @return the modelId
	 * @see #modelId
	 */
	public Integer getModelId() {
		return modelId;
	}

	/**
	 * @param modelId
	 *          the modelId to set
	 * @see #modelId
	 */
	public void setModelId(Integer modelId) {
		this.modelId = modelId;
	}

	/**
	 * @return the type
	 * @see #type
	 */
	public String getType() {
		return type;
	}

	/**
	 * @param type
	 *          the type to set
	 * @see #type
	 */
	public void setType(String type) {
		this.type = type;
	}

	/**
	 * @return the icon
	 * @see #icon
	 */
	public String getIcon() {
		return icon;
	}

	/**
	 * @param icon
	 *          the icon to set
	 * @see #icon
	 */
	public void setIcon(String icon) {
		this.icon = icon;
	}

	@Override
	public String toString() {
		String result = "[" + this.getClass().getSimpleName() + "]\t" + //
				"objectId:" + objectId + "\t" + //
				"modelId:" + modelId + "\t" + //
				"type:" + type + "\t";
		return result;
	}
}
