package lcsb.mapviewer.services.search.data;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.Species;
import lcsb.mapviewer.services.search.IHeavyView;
import lcsb.mapviewer.services.view.AnnotationView;

/**
 * Heavy {@link LightAliasView view} for alias representing data that should be
 * sent to client.
 * 
 * @author Piotr Gawron
 *
 */
public class FullAliasView extends LightAliasView implements IHeavyView {

	/**
	 * Default class logger.
	 */
	@SuppressWarnings("unused")
	private static Logger logger = Logger.getLogger(FullAliasView.class);

	/**
	 * This class represents posttransational modification of the element.
	 * 
	 * @author Piotr Gawron
	 * 
	 */
	public static class PosttransationalModification implements Serializable {

		/**
		 * 
		 */
		private static final long	serialVersionUID = 1L;

		/**
		 * Name of the modification.
		 */
		private String						name;

		/**
		 * Position where the modification occurred.
		 */
		private String						position;

		/**
		 * @return the name
		 * @see #name
		 */
		public String getName() {
			return name;
		}

		/**
		 * @param name
		 *          the name to set
		 * @see #name
		 */
		public void setName(String name) {
			this.name = name;
		}

		/**
		 * @return the position
		 * @see #position
		 */
		public String getPosition() {
			return position;
		}

		/**
		 * @param position
		 *          the position to set
		 * @see #position
		 */
		public void setPosition(String position) {
			this.position = position;
		}
	}

	/**
	 * 
	 */
	private static final long		 serialVersionUID			= 1L;

	/**
	 * Description.
	 * 
	 * @see Element#notes
	 */
	private String							 notes;

	/**
	 * Type of the alias.
	 * 
	 * @see Element#getStringType()
	 */
	private String							 type;

	/**
	 * Symbol of the element.
	 * 
	 * @see Element#symbol
	 */
	private String							 symbol;

	/**
	 * Full name of the element.
	 * 
	 * @see Element#fullName
	 */
	private String							 fullName;

	/**
	 * Abbreviation of the element.
	 * 
	 * @see Element#abbreviation
	 */
	private String							 abbreviation;

	/**
	 * Formula describing the element.
	 * 
	 * @see Element#formula
	 */
	private String							 formula;

	/**
	 * Charge of the element.
	 * 
	 * @see Species#charge
	 */
	private Integer							 charge;

	/**
	 * Initial amount of species.
	 * 
	 * @see Species#initialAmount
	 */
	private Integer							 initialAmount				= null;

	private String							 visibilityLevel			= null;

	/**
	 * Initial concentration of species.
	 * 
	 * @see Species#initialConcentration
	 */
	private Integer							 initialConcentration	= null;

	/**
	 * Is only substance units allowed in a species.
	 * 
	 * @see Species#onlySubstanceUnits
	 */
	private Boolean							 onlySubstanceUnits		= null;

	/**
	 * Name of the element.
	 * 
	 * @see Element#name
	 */
	private String							 name									= "";

	/**
	 * Name of the {@link Element#model}.
	 */
	private String							 modelName;

	/**
	 * Name of the {@link Element#submodel}.
	 */
	private String							 submodelName;

	/**
	 * Id of the {@link Element#submodel}.
	 */
	private String							 submodelId;

	/**
	 * List of synonyms of the element.
	 * 
	 * @see Element#synonyms
	 */
	private List<String>				 synonyms							= new ArrayList<>();

	/**
	 * List of former symbols used by the entity represented by this element.
	 * 
	 * @see Element#formerSymbols
	 */
	private List<String>				 formerSymbols				= new ArrayList<>();

	/**
	 * List of references for this element.
	 * 
	 * @see Element#miriamData
	 */
	private List<AnnotationView> references						= new ArrayList<>();

	/**
	 * List of other unstructurized feuters, specific for some subclases of Alias.
	 * 
	 */
	private Map<String, Object>	 other								= new HashMap<>();

	/**
	 * Default constructor.
	 * 
	 * @param element
	 *          original object from which this element is created
	 * @param icon
	 *          string with icon that should be used to visualize this alias
	 */
	protected FullAliasView(Element element, String icon) {
		super(element, icon);
		setModelName(element.getModel().getName());
		if (element != null) {
			setAbbreviation(element.getAbbreviation());
			addFormerSymbols(element.getFormerSymbols());
			setFormula(element.getFormula());
			setFullName(element.getFullName());
			setName(element.getName());
			setNotes(element.getNotes());
			setSymbol(element.getSymbol());
			addSynonyms(element.getSynonyms());
			setType(element.getStringType());
			setVisibilityLevel(element.getVisibilityLevel());
			if (element instanceof Species) {
				Species species = (Species) element;
				setCharge(species.getCharge());
				setInitialAmount(species.getInitialAmount());
				setInitialConcentration(species.getInitialConcentration());
				setOnlySubstanceUnits(species.getOnlySubstanceUnits());
				if (element.getSubmodel() != null && element.getSubmodel().getSubmodel() != null) {
					setSubmodelName(element.getSubmodel().getSubmodel().getName());
					setSubmodelId(element.getSubmodel().getSubmodel().getId() + "");
				}
			}
		}
	}

	/**
	 * Constructor that should be used only by the deserialization method.
	 */
	protected FullAliasView() {
	}

	/**
	 * Adds elements to {@link #synonyms} list.
	 * 
	 * @param synonyms2
	 *          list of elements to add
	 */
	private void addSynonyms(List<String> synonyms2) {
		synonyms.addAll(synonyms2);
	}

	/**
	 * Adds elements to {@link #formerSymbols} list.
	 * 
	 * @param formerSymbols2
	 *          list of elements to add
	 */
	private void addFormerSymbols(List<String> formerSymbols2) {
		formerSymbols.addAll(formerSymbols2);
	}

	/**
	 * @return the notes
	 * @see #notes
	 */
	public String getNotes() {
		return notes;
	}

	/**
	 * @param notes
	 *          the notes to set
	 * @see #notes
	 */
	public void setNotes(String notes) {
		this.notes = notes;
	}

	/**
	 * @return the symbol
	 * @see #symbol
	 */
	public String getSymbol() {
		return symbol;
	}

	/**
	 * @param symbol
	 *          the symbol to set
	 * @see #symbol
	 */
	public void setSymbol(String symbol) {
		this.symbol = symbol;
	}

	/**
	 * @return the fullName
	 * @see #fullName
	 */
	public String getFullName() {
		return fullName;
	}

	/**
	 * @param fullName
	 *          the fullName to set
	 * @see #fullName
	 */
	public void setFullName(String fullName) {
		this.fullName = fullName;
	}

	/**
	 * @return the abbreviation
	 * @see #abbreviation
	 */
	public String getAbbreviation() {
		return abbreviation;
	}

	/**
	 * @param abbreviation
	 *          the abbreviation to set
	 * @see #abbreviation
	 */
	public void setAbbreviation(String abbreviation) {
		this.abbreviation = abbreviation;
	}

	/**
	 * @return the formula
	 * @see #formula
	 */
	public String getFormula() {
		return formula;
	}

	/**
	 * @param formula
	 *          the formula to set
	 * @see #formula
	 */
	public void setFormula(String formula) {
		this.formula = formula;
	}

	/**
	 * @return the name
	 * @see #name
	 */
	public String getName() {
		return name;
	}

	/**
	 * @param name
	 *          the name to set
	 * @see #name
	 */
	public void setName(String name) {
		this.name = name;
	}

	/**
	 * @return the synonyms
	 * @see #synonyms
	 */
	public List<String> getSynonyms() {
		return synonyms;
	}

	/**
	 * @param synonyms
	 *          the synonyms to set
	 * @see #synonyms
	 */
	public void setSynonyms(List<String> synonyms) {
		this.synonyms = synonyms;
	}

	/**
	 * @return the formerSymbols
	 * @see #formerSymbols
	 */
	public List<String> getFormerSymbols() {
		return formerSymbols;
	}

	/**
	 * @param formerSymbols
	 *          the formerSymbols to set
	 * @see #formerSymbols
	 */
	public void setFormerSymbols(List<String> formerSymbols) {
		this.formerSymbols = formerSymbols;
	}

	/**
	 * @return the references
	 * @see #references
	 */
	public List<AnnotationView> getReferences() {
		return references;
	}

	/**
	 * @param references
	 *          the references to set
	 * @see #references
	 */
	public void setReferences(List<AnnotationView> references) {
		this.references = references;
	}

	/**
	 * @return the other
	 * @see #other
	 */
	public Map<String, Object> getOther() {
		return other;
	}

	/**
	 * Returns other (additional element) by the name.
	 * 
	 * @param key
	 *          key identifing the feature.
	 * @return other (additional element) by the name
	 */
	public Object getOther(String key) {
		return other.get(key);
	}

	/**
	 * @param other
	 *          the other to set
	 * @see #other
	 */
	public void setOther(Map<String, Object> other) {
		this.other = other;
	}

	/**
	 * @return the type
	 * @see #type
	 */
	public String getType() {
		return type;
	}

	/**
	 * @param type
	 *          the type to set
	 * @see #type
	 */
	public void setType(String type) {
		this.type = type;
	}

	/**
	 * Adds elements to {@link #references} list.
	 * 
	 * @param newReferences
	 *          list of elements to add
	 */
	public void addReferences(List<AnnotationView> newReferences) {
		references.addAll(newReferences);
	}

	/**
	 * Addsother (additional element) by the name.
	 * 
	 * @param key
	 *          name of the feature
	 * @param value
	 *          object representing data of the feature
	 */
	public void addOther(String key, Object value) {
		other.put(key, value);
	}

	/**
	 * @return the modelName
	 * @see #modelName
	 */
	public String getModelName() {
		return modelName;
	}

	/**
	 * @param modelName
	 *          the modelName to set
	 * @see #modelName
	 */
	public void setModelName(String modelName) {
		this.modelName = modelName;
	}

	/**
	 * @return the charge
	 * @see #charge
	 */
	public Integer getCharge() {
		return charge;
	}

	/**
	 * @param charge
	 *          the charge to set
	 * @see #charge
	 */
	public void setCharge(Integer charge) {
		this.charge = charge;
	}

	/**
	 * @return the initialAmount
	 * @see #initialAmount
	 */
	public Integer getInitialAmount() {
		return initialAmount;
	}

	/**
	 * @param initialAmount
	 *          the initialAmount to set
	 * @see #initialAmount
	 */
	public void setInitialAmount(Integer initialAmount) {
		this.initialAmount = initialAmount;
	}

	/**
	 * @return the initialConcentration
	 * @see #initialConcentration
	 */
	public Integer getInitialConcentration() {
		return initialConcentration;
	}

	/**
	 * @param initialConcentration
	 *          the initialConcentration to set
	 * @see #initialConcentration
	 */
	public void setInitialConcentration(Integer initialConcentration) {
		this.initialConcentration = initialConcentration;
	}

	/**
	 * @return the onlySubstanceUnits
	 * @see #onlySubstanceUnits
	 */
	public Boolean getOnlySubstanceUnits() {
		return onlySubstanceUnits;
	}

	/**
	 * @param onlySubstanceUnits
	 *          the onlySubstanceUnits to set
	 * @see #onlySubstanceUnits
	 */
	public void setOnlySubstanceUnits(Boolean onlySubstanceUnits) {
		this.onlySubstanceUnits = onlySubstanceUnits;
	}

	/**
	 * @return the submodelName
	 * @see #submodelName
	 */
	public String getSubmodelName() {
		return submodelName;
	}

	/**
	 * @param submodelName
	 *          the submodelName to set
	 * @see #submodelName
	 */
	public void setSubmodelName(String submodelName) {
		this.submodelName = submodelName;
	}

	/**
	 * @return the submodelId
	 * @see #submodelId
	 */
	public String getSubmodelId() {
		return submodelId;
	}

	/**
	 * @param submodelId
	 *          the submodelId to set
	 * @see #submodelId
	 */
	public void setSubmodelId(String submodelId) {
		this.submodelId = submodelId;
	}

	/**
	 * @return the visibilityLevel
	 * @see #visibilityLevel
	 */
	public String getVisibilityLevel() {
		return visibilityLevel;
	}

	/**
	 * @param visibilityLevel
	 *          the visibilityLevel to set
	 * @see #visibilityLevel
	 */
	public void setVisibilityLevel(String visibilityLevel) {
		this.visibilityLevel = visibilityLevel;
	}
}
