package lcsb.mapviewer.services.view;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

import lcsb.mapviewer.common.comparator.IntegerComparator;
import lcsb.mapviewer.model.Project;

/**
 * This class is a client view that represents {@link Project} class. Right now
 * this class highly entangles {@link Project} and
 * {@link lcsb.mapviewer.model.map.model.db.model.map.Model Model}. In the
 * future it should separate it.
 * 
 * @author Piotr Gawron
 * 
 */
public class ProjectView extends AbstractView<Project> implements Comparable<ProjectView>, Serializable {

	/**
	 * 
	 */
	private static final long				serialVersionUID = 1L;

	/**
	 * Identifier of the model that is represented in this project view (this must
	 * be refactorized in the future).
	 */
	private Integer									modelId;

	/**
	 * List of layouts in the model (identified by {@link #modelId}).
	 */
	private List<LayoutView>				layouts					 = new ArrayList<LayoutView>();

	/**
	 * {@link Project#projectId Identifier} of the project.
	 */
	private String									projectId;

	/**
	 * {@link Project#name Name} of the project.
	 */
	private String									projectName;

	/**
	 * {@link Project#name newDiseaseName} of the project.
	 */
	private String									newDiseaseName;

	/**
	 * {@link Project#organism newOrganismName} of the project.
	 */
	private String									newOrganismName;

	/**
	 * {@link Project#disease Disease} of the project.
	 */
	private AnnotationView					projectDiseaseLink;

	/**
	 * {@link Project#disease Disease name} of the project.
	 */
	private String									diseaseName			 = "N/A";

	/**
	 * {@link Project#organism Organism} of the project.
	 */
	private AnnotationView					organismLink;

	/**
	 * Email used for notification obut the events connected to this project.
	 */
	private String									notifyEmail;

	/**
	 * Status of the project processing.
	 */
	private String									status;

	/**
	 * Errors that appeard in the project.
	 */
	private String									errors;

	/**
	 * Progress of the project processng (value between 0..100).
	 */
	private Integer									progress;

	/**
	 * Short list of warnings that appeard during model creation.
	 */
	private String									warningsPreview;

	/**
	 * List of warnings that appeard during model creation.
	 */
	private List<String>						warnings;

	/**
	 * Description of the model.
	 */
	private String									description;
	/**
	 * Version of the model.
	 */
	private String									version;

	/**
	 * "true" if input file from which project was created is available, "false"
	 * otherwise.
	 */
	private String									inputDataAvailable;

	/**
	 * List of data mining sets used to gernerate suggested connections.
	 */
	private List<DataMiningSetView>	dataMiningSets	 = new ArrayList<DataMiningSetView>();

	/**
	 * Default constructor.
	 * 
	 * @param project
	 *          {@link Project} for which this view is created
	 */
	protected ProjectView(Project project) {
		super(project);
	}

	/**
	 * Default constructor. Should be used only for deserialization.
	 */
	protected ProjectView() {
	}

	/**
	 * Constructor that creates copy of the {@link ProjectView}.
	 * 
	 * @param row
	 *          original object
	 */
	public ProjectView(ProjectView row) {
		super(null);
		setIdObject(row.getIdObject());
		modelId = row.getModelId();
		layouts.addAll(row.getLayouts());
		projectId = row.getProjectId();
		projectName = row.getProjectName();
		newDiseaseName = row.getNewDiseaseName();
		newOrganismName = row.getNewOrganismName();
		projectDiseaseLink = row.getProjectDiseaseLink();
		organismLink = row.getOrganismLink();
		notifyEmail = row.getNotifyEmail();
		status = row.getStatus();
		progress = row.getProgress();
		description = row.getDescription();
		version = row.getVersion();
		errors = row.getErrors();
		warningsPreview = row.getWarningsPreview();
		warnings = row.getWarnings();
		for (DataMiningSetView view : row.getDataMiningSets()) {
			dataMiningSets.add(new DataMiningSetView(view));
		}
		inputDataAvailable = row.getInputDataAvailable();
	}

	@Override
	public int compareTo(ProjectView o) {
		if (o == null) {
			new IntegerComparator().compare(getIdObject(), null);
		}
		return new IntegerComparator().compare(getIdObject(), o.getIdObject());
	}

	/**
	 * @return the modelId
	 * @see #modelId
	 */
	public Integer getModelId() {
		return modelId;
	}

	/**
	 * @param modelId
	 *          the modelId to set
	 * @see #modelId
	 */
	public void setModelId(Integer modelId) {
		this.modelId = modelId;
	}

	/**
	 * @return the layouts
	 * @see #layouts
	 */
	public List<LayoutView> getLayouts() {
		return layouts;
	}

	/**
	 * @param layouts
	 *          the layouts to set
	 * @see #layouts
	 */
	public void setLayouts(List<LayoutView> layouts) {
		this.layouts = layouts;
	}

	/**
	 * @return the name
	 * @see #projectId
	 */
	public String getProjectId() {
		return projectId;
	}

	/**
	 * @param projectId
	 *          project identifier
	 * @see #projectId
	 */
	public void setProjectId(String projectId) {
		this.projectId = projectId;
	}

	/**
	 * @return the notifyEmail
	 * @see #notifyEmail
	 */
	public String getNotifyEmail() {
		return notifyEmail;
	}

	/**
	 * @param notifyEmail
	 *          the notifyEmail to set
	 * @see #notifyEmail
	 */
	public void setNotifyEmail(String notifyEmail) {
		this.notifyEmail = notifyEmail;
	}

	/**
	 * @return the status
	 * @see #status
	 */
	public String getStatus() {
		return status;
	}

	/**
	 * @param status
	 *          the status to set
	 * @see #status
	 */
	public void setStatus(String status) {
		this.status = status;
	}

	/**
	 * @return the progress
	 * @see #progress
	 */
	public Integer getProgress() {
		return progress;
	}

	/**
	 * @param progress
	 *          the progress to set
	 * @see #progress
	 */
	public void setProgress(Integer progress) {
		this.progress = progress;
	}

	/**
	 * @return the description
	 * @see #description
	 */
	public String getDescription() {
		return description;
	}

	/**
	 * @param description
	 *          the description to set
	 * @see #description
	 */
	public void setDescription(String description) {
		this.description = description;
	}

	/**
	 * @return the version
	 * @see #version
	 */
	public String getVersion() {
		return version;
	}

	/**
	 * @param version
	 *          the version to set
	 * @see #version
	 */
	public void setVersion(String version) {
		this.version = version;
	}

	/**
	 * Adds layout to the view.
	 * 
	 * @param row
	 *          layout to add
	 */
	public void addLayout(LayoutView row) {
		layouts.add(row);
	}

	/**
	 * @return the warnings
	 * @see #warnings
	 */
	public List<String> getWarnings() {
		return warnings;
	}

	/**
	 * @param warnings
	 *          the warnings to set
	 * @see #warnings
	 */
	public void setWarnings(List<String> warnings) {
		this.warnings = warnings;
	}

	/**
	 * @return the errors
	 * @see #errors
	 */
	public String getErrors() {
		return errors;
	}

	/**
	 * @param errors
	 *          the errors to set
	 * @see #errors
	 */
	public void setErrors(String errors) {
		this.errors = errors;
	}

	/**
	 * @return the dataMiningSets
	 * @see #dataMiningSets
	 */
	public List<DataMiningSetView> getDataMiningSets() {
		return dataMiningSets;
	}

	/**
	 * @param dataMiningSets
	 *          the dataMiningSets to set
	 * @see #dataMiningSets
	 */
	public void setDataMiningSets(List<DataMiningSetView> dataMiningSets) {
		this.dataMiningSets = dataMiningSets;
	}

	/**
	 * Adds {@link DataMiningSetView} to {@link #dataMiningSets}.
	 * 
	 * @param dataMiningSetView
	 *          object to add
	 */
	public void addDataMiningSet(DataMiningSetView dataMiningSetView) {
		dataMiningSets.add(dataMiningSetView);
	}

	/**
	 * @return the projectName
	 * @see #projectName
	 */
	public String getProjectName() {
		return projectName;
	}

	/**
	 * @param projectName
	 *          the projectName to set
	 * @see #projectName
	 */
	public void setProjectName(String projectName) {
		this.projectName = projectName;
	}

	/**
	 * @return project disease link
	 */
	public AnnotationView getProjectDiseaseLink() {
		return projectDiseaseLink;
	}

	/**
	 * @param projectDiseaseLink
	 *          the link for the disease.
	 */
	public void setProjectDiseaseLink(AnnotationView projectDiseaseLink) {
		this.projectDiseaseLink = projectDiseaseLink;
	}

	/**
	 * @return this is used to update or add different disease association.
	 */
	public String getNewDiseaseName() {
		return newDiseaseName;
	}

	/**
	 * @param newDiseaseName
	 *          this is used to update or add different disease association.
	 */
	public void setNewDiseaseName(String newDiseaseName) {
		this.newDiseaseName = newDiseaseName;
	}

	/**
	 * @return the diseaseName
	 * @see #diseaseName
	 */
	public String getDiseaseName() {
		return diseaseName;
	}

	/**
	 * @param diseaseName
	 *          the diseaseName to set
	 * @see #diseaseName
	 */
	public void setDiseaseName(String diseaseName) {
		this.diseaseName = diseaseName;
	}

	/**
	 * @return the inputDataAvailable
	 * @see #inputDataAvailable
	 */
	public String getInputDataAvailable() {
		return inputDataAvailable;
	}

	/**
	 * @param inputDataAvailable
	 *          the inputDataAvailable to set
	 * @see #inputDataAvailable
	 */
	public void setInputDataAvailable(String inputDataAvailable) {
		this.inputDataAvailable = inputDataAvailable;
	}

	/**
	 * @param value
	 *          the inputDataAvailable to set
	 * @see #inputDataAvailable
	 */
	public void setInputDataAvailable(boolean value) {
		if (value) {
			this.inputDataAvailable = "true";
		} else {
			this.inputDataAvailable = "false";
		}

	}

	/**
	 * @return the organismLink
	 * @see #organismLink
	 */
	public AnnotationView getOrganismLink() {
		return organismLink;
	}

	/**
	 * @param organismLink the organismLink to set
	 * @see #organismLink
	 */
	public void setOrganismLink(AnnotationView organismLink) {
		this.organismLink = organismLink;
	}

	/**
	 * @return the newOrganismName
	 * @see #newOrganismName
	 */
	public String getNewOrganismName() {
		return newOrganismName;
	}

	/**
	 * @param newOrganismName the newOrganismName to set
	 * @see #newOrganismName
	 */
	public void setNewOrganismName(String newOrganismName) {
		this.newOrganismName = newOrganismName;
	}

	/**
	 * @return the warningsPreview
	 * @see #warningsPreview
	 */
	public String getWarningsPreview() {
		return warningsPreview;
	}

	/**
	 * @param warningsPreview the warningsPreview to set
	 * @see #warningsPreview
	 */
	public void setWarningsPreview(String warningsPreview) {
		this.warningsPreview = warningsPreview;
	}

}
